@@ -1,4 +1,3 @@
-#include <errno.h>
 #include <poll.h>
 #include <limits.h>
 #include <setjmp.h>
@@ -9,6 +8,7 @@
 #include <arpa/inet.h>
 #include <sys/types.h>
 #include <sys/time.h>
+#include <sys/sysinfo.h>
 #include <dirent.h>
 #include <fcntl.h>
 #include <unistd.h>		/* for getopts */
@@ -30,6 +30,7 @@ static unsigned int default_caps = CAP_DHCP | CAP_BOOTP | CAP_RARP;
 static int loop_timeout = -1;
 static int configured;
 static int bringup_first = 0;
+static int n_devices = 0;
 
 /* DHCP vendor class identifier */
 char vendor_class_identifier[260];
@@ -45,6 +46,20 @@ struct state {
 	struct state *next;
 };
 
+/* #define PROTO_x : for uint8_t proto of struct netdev */
+struct protoinfo {
+	char *name;
+} protoinfos[] = {
+#define PROTO_NONE  0
+	{"none"},
+#define PROTO_BOOTP 1
+	{"bootp"},
+#define PROTO_DHCP  2
+	{"dhcp"},
+#define PROTO_RARP  3
+	{"rarp"}
+};
+
 static inline const char *my_inet_ntoa(uint32_t addr)
 {
 	struct in_addr a;
@@ -56,14 +71,38 @@ static inline const char *my_inet_ntoa(uint32_t addr)
 
 static void print_device_config(struct netdev *dev)
 {
-	printf("IP-Config: %s complete (from %s):\n", dev->name,
-	       my_inet_ntoa(dev->serverid ? dev->serverid : dev->ip_server));
-	printf(" address: %-16s ", my_inet_ntoa(dev->ip_addr));
+	int dns0_spaces;
+	int dns1_spaces;
+	printf("IP-Config: %s complete", dev->name);
+	if (dev->proto == PROTO_BOOTP || dev->proto == PROTO_DHCP)
+		printf(" (%s from %s)", protoinfos[dev->proto].name,
+		       my_inet_ntoa(dev->serverid ?
+				    dev->serverid : dev->ip_server));
+
+	printf(":\n address: %-16s ", my_inet_ntoa(dev->ip_addr));
 	printf("broadcast: %-16s ", my_inet_ntoa(dev->ip_broadcast));
 	printf("netmask: %-16s\n", my_inet_ntoa(dev->ip_netmask));
-	printf(" gateway: %-16s ", my_inet_ntoa(dev->ip_gateway));
-	printf("dns0     : %-16s ", my_inet_ntoa(dev->ip_nameserver[0]));
-	printf("dns1   : %-16s\n", my_inet_ntoa(dev->ip_nameserver[1]));
+	if (dev->routes != NULL) {
+		struct route *cur;
+		char *delim = "";
+		printf(" routes :");
+		for (cur = dev->routes; cur != NULL; cur = cur->next) {
+			printf("%s %s/%u", delim, my_inet_ntoa(cur->subnet), cur->netmask_width);
+			if (cur->gateway != 0) {
+				printf(" via %s", my_inet_ntoa(cur->gateway));
+			}
+			delim = ",";
+		}
+		printf("\n");
+		dns0_spaces = 3;
+		dns1_spaces = 5;
+	} else {
+		printf(" gateway: %-16s", my_inet_ntoa(dev->ip_gateway));
+		dns0_spaces = 5;
+		dns1_spaces = 3;
+	}
+	printf(" dns0%*c: %-16s", dns0_spaces, ' ', my_inet_ntoa(dev->ip_nameserver[0]));
+	printf(" dns1%*c: %-16s\n", dns1_spaces, ' ', my_inet_ntoa(dev->ip_nameserver[1]));
 	if (dev->hostname[0])
 		printf(" host   : %-64s\n", dev->hostname);
 	if (dev->dnsdomainname[0])
@@ -87,8 +126,8 @@ static void configure_device(struct netdev *dev)
 	if (netdev_setaddress(dev))
 		printf("IP-Config: failed to set addresses on %s\n",
 		       dev->name);
-	if (netdev_setdefaultroute(dev))
-		printf("IP-Config: failed to set default route on %s\n",
+	if (netdev_setroutes(dev))
+		printf("IP-Config: failed to set routes on %s\n",
 		       dev->name);
 	if (dev->hostname[0] &&
 			sethostname(dev->hostname, strlen(dev->hostname)))
@@ -96,30 +135,87 @@ static void configure_device(struct netdev *dev)
 			dev->hostname, dev->name);
 }
 
+/*
+ * Escape shell varialbes in git style:
+ * Always start with a single quote ('), then leave all characters
+ * except ' and ! unchanged.
+ */
+static void write_option(FILE *f, const char *name, const char *chr)
+{
+
+	fprintf(f, "%s='", name);
+	while (*chr) {
+		switch (*chr) {
+		case '!':
+		case '\'':
+			fprintf(f, "'\\%c'", *chr);
+			break;
+		default:
+			fprintf(f, "%c", *chr);
+			break;
+		}
+		++chr;
+	}
+	fprintf(f, "'\n");
+}
+
 static void dump_device_config(struct netdev *dev)
 {
 	char fn[40];
 	FILE *f;
+	/*
+	 * char UINT64_MAX[] = "18446744073709551615";
+	 * sizeof(UINT64_MAX)==21
+	 */
+	char buf21[21];
+	const char path[] = "/run/";
 
-	snprintf(fn, sizeof(fn), "/tmp/net-%s.conf", dev->name);
+	snprintf(fn, sizeof(fn), "%snet-%s.conf", path, dev->name);
 	f = fopen(fn, "w");
 	if (f) {
-		fprintf(f, "DEVICE=%s\n", dev->name);
-		fprintf(f, "IPV4ADDR=%s\n", my_inet_ntoa(dev->ip_addr));
-		fprintf(f, "IPV4BROADCAST=%s\n",
-			my_inet_ntoa(dev->ip_broadcast));
-		fprintf(f, "IPV4NETMASK=%s\n", my_inet_ntoa(dev->ip_netmask));
-		fprintf(f, "IPV4GATEWAY=%s\n", my_inet_ntoa(dev->ip_gateway));
-		fprintf(f, "IPV4DNS0=%s\n",
-			my_inet_ntoa(dev->ip_nameserver[0]));
-		fprintf(f, "IPV4DNS1=%s\n",
-			my_inet_ntoa(dev->ip_nameserver[1]));
-		fprintf(f, "HOSTNAME=%s\n", dev->hostname);
-		fprintf(f, "DNSDOMAIN=%s\n", dev->dnsdomainname);
-		fprintf(f, "NISDOMAIN=%s\n", dev->nisdomainname);
-		fprintf(f, "ROOTSERVER=%s\n", my_inet_ntoa(dev->ip_server));
-		fprintf(f, "ROOTPATH=%s\n", dev->bootpath);
-		fprintf(f, "filename=\"%s\"\n", dev->filename);
+		write_option(f, "DEVICE", dev->name);
+		write_option(f, "PROTO", protoinfos[dev->proto].name);
+		write_option(f, "IPV4ADDR",
+				my_inet_ntoa(dev->ip_addr));
+		write_option(f, "IPV4BROADCAST",
+				my_inet_ntoa(dev->ip_broadcast));
+		write_option(f, "IPV4NETMASK",
+				my_inet_ntoa(dev->ip_netmask));
+		if (dev->routes != NULL) {
+			/* Use 6 digits to encode the index */
+			char key[23];
+			char value[19];
+			int i = 0;
+			struct route *cur;
+			for (cur = dev->routes; cur != NULL; cur = cur->next) {
+				snprintf(key, sizeof(key), "IPV4ROUTE%iSUBNET", i);
+				snprintf(value, sizeof(value), "%s/%u", my_inet_ntoa(cur->subnet), cur->netmask_width);
+				write_option(f, key, value);
+				snprintf(key, sizeof(key), "IPV4ROUTE%iGATEWAY", i);
+				write_option(f, key, my_inet_ntoa(cur->gateway));
+				i++;
+			}
+		} else {
+			write_option(f, "IPV4GATEWAY",
+					my_inet_ntoa(dev->ip_gateway));
+		}
+		write_option(f, "IPV4DNS0",
+				my_inet_ntoa(dev->ip_nameserver[0]));
+		write_option(f, "IPV4DNS1",
+				my_inet_ntoa(dev->ip_nameserver[1]));
+		write_option(f, "HOSTNAME",  dev->hostname);
+		write_option(f, "DNSDOMAIN", dev->dnsdomainname);
+		write_option(f, "NISDOMAIN", dev->nisdomainname);
+		write_option(f, "ROOTSERVER",
+				my_inet_ntoa(dev->ip_server));
+		write_option(f, "ROOTPATH", dev->bootpath);
+		write_option(f, "filename", dev->filename);
+		sprintf(buf21, "%ld", (long)dev->uptime);
+		write_option(f, "UPTIME", buf21);
+		sprintf(buf21, "%u", (unsigned int)dev->dhcpleasetime);
+		write_option(f, "DHCPLEASETIME", buf21);
+		write_option(f, "DOMAINSEARCH", dev->domainsearch == NULL ?
+			     "" : dev->domainsearch);
 		fclose(f);
 	}
 }
@@ -149,19 +245,19 @@ static void postprocess_device(struct netdev *dev)
 		printf("IP-Config: %s guessed broadcast address %s\n",
 		       dev->name, my_inet_ntoa(dev->ip_broadcast));
 	}
-	if (dev->ip_nameserver[0] == INADDR_ANY) {
-		dev->ip_nameserver[0] = dev->ip_server;
-		printf("IP-Config: %s guessed nameserver address %s\n",
-		       dev->name, my_inet_ntoa(dev->ip_nameserver[0]));
-	}
 }
 
 static void complete_device(struct netdev *dev)
 {
+	struct sysinfo info;
+
+	if (!sysinfo(&info))
+		dev->uptime = info.uptime;
 	postprocess_device(dev);
 	configure_device(dev);
 	dump_device_config(dev);
 	print_device_config(dev);
+	packet_close(dev);
 
 	++configured;
 
@@ -171,7 +267,7 @@ static void complete_device(struct netdev *dev)
 
 /*
  * Returns:
- *  0 = Not handled, the packet is still in the queue
+ *  0 = Not handled, try again later
  *  1 = Handled
  */
 static int process_receive_event(struct state *s, time_t now)
@@ -179,14 +275,23 @@ static int process_receive_event(struct state *s, time_t now)
 	int handled = 1;
 
 	switch (s->state) {
+	case DEVST_ERROR:
+		return 0; /* Not handled */
+	case DEVST_COMPLETE:
+		return 0; /* Not handled as already configured */
+
 	case DEVST_BOOTP:
 		s->restart_state = DEVST_BOOTP;
 		switch (bootp_recv_reply(s->dev)) {
 		case -1:
 			s->state = DEVST_ERROR;
 			break;
+		case 0:
+			handled = 0;
+			break;
 		case 1:
 			s->state = DEVST_COMPLETE;
+			s->dev->proto = PROTO_BOOTP;
 			dprintf("\n   bootp reply\n");
 			break;
 		}
@@ -198,6 +303,9 @@ static int process_receive_event(struct state *s, time_t now)
 		case -1:
 			s->state = DEVST_ERROR;
 			break;
+		case 0:
+			handled = 0;
+			break;
 		case DHCPOFFER:	/* Offer received */
 			s->state = DEVST_DHCPREQ;
 			dhcp_send_request(s->dev);
@@ -211,8 +319,12 @@ static int process_receive_event(struct state *s, time_t now)
 		case -1:	/* error */
 			s->state = DEVST_ERROR;
 			break;
+		case 0:
+			handled = 0;
+			break;
 		case DHCPACK:	/* ACK received */
 			s->state = DEVST_COMPLETE;
+			s->dev->proto = PROTO_DHCP;
 			break;
 		case DHCPNAK:	/* NAK received */
 			s->state = DEVST_DHCPDISC;
@@ -297,49 +409,39 @@ struct netdev *ifaces;
 
 /*
  * Returns:
- *  0 = Error, packet not received or discarded
+ *  0 = No dhcp/bootp packet was received
  *  1 = A packet was received and handled
  */
-static int do_pkt_recv(int pkt_fd, time_t now)
+static int do_pkt_recv(int nr, struct pollfd *fds, time_t now)
 {
-	int ifindex, ret;
+	int i, ret = 0;
 	struct state *s;
 
-	ret = packet_peek(&ifindex);
-	if (ret == 0)
-		return ret;
-
-	for (s = slist; s; s = s->next) {
-		if (s->dev->ifindex == ifindex) {
-			ret = process_receive_event(s, now);
-			break;
+	for (i = 0, s = slist; s && nr; s = s->next, i++) {
+		if (fds[i].revents & POLLRDNORM) {
+			ret |= process_receive_event(s, now);
+			nr--;
 		}
 	}
-
-	if (ret == 0)
-		packet_discard();
-
 	return ret;
 }
 
 static int loop(void)
 {
-#define NR_FDS	1
-	struct pollfd fds[NR_FDS];
+	struct pollfd *fds;
 	struct state *s;
-	int pkt_fd;
-	int nr = 0;
+	int i, nr = 0, rc = 0;
 	struct timeval now, prev;
 	time_t start;
 
-	pkt_fd = packet_open();
-	if (pkt_fd == -1) {
-		perror("packet_open");
-		return -1;
+	fds = malloc(sizeof(struct pollfd) * n_devices);
+	if (!fds) {
+		fprintf(stderr, "malloc failed\n");
+		rc = -1;
+		goto bail;
 	}
 
-	fds[0].fd = pkt_fd;
-	fds[0].events = POLLRDNORM;
+	memset(fds, 0, sizeof(*fds));
 
 	gettimeofday(&now, NULL);
 	start = now.tv_sec;
@@ -350,9 +452,12 @@ static int loop(void)
 		int timeout_ms;
 		int x;
 
-		for (s = slist; s; s = s->next) {
+		for (i = 0, s = slist; s; s = s->next, i++) {
 			dprintf("%s: state = %d\n", s->dev->name, s->state);
 
+			fds[i].fd = s->dev->pkt_fd;
+			fds[i].events = POLLRDNORM;
+
 			if (s->state == DEVST_COMPLETE) {
 				done++;
 				continue;
@@ -380,19 +485,18 @@ static int loop(void)
 			if (timeout_ms <= 0)
 				timeout_ms = 100;
 
-			nr = poll(fds, NR_FDS, timeout_ms);
+			nr = poll(fds, n_devices, timeout_ms);
 			prev = now;
 			gettimeofday(&now, NULL);
 
-			if ((fds[0].revents & POLLRDNORM)) {
-				if (do_pkt_recv(pkt_fd, now.tv_sec) == 1)
-					break;
-			}
+			if ((nr > 0) && do_pkt_recv(nr, fds, now.tv_sec))
+				break;
 
 			if (loop_timeout >= 0 &&
 			    now.tv_sec - start >= loop_timeout) {
 				printf("IP-Config: no response after %d "
 				       "secs - giving up\n", loop_timeout);
+				rc = -1;
 				goto bail;
 			}
 
@@ -404,10 +508,10 @@ static int loop(void)
 			timeout_ms -= delta_ms;
 		}
 	}
-      bail:
-	packet_close();
-
-	return 0;
+bail:
+	if (fds)
+		free(fds);
+	return rc;
 }
 
 static int add_one_dev(struct netdev *dev)
@@ -425,21 +529,22 @@ static int add_one_dev(struct netdev *dev)
 	/*
 	 * Select the state that we start from.
 	 */
-	if (dev->caps & CAP_DHCP && dev->ip_addr == INADDR_ANY) {
+	if (dev->caps & CAP_DHCP && dev->ip_addr == INADDR_ANY)
 		state->restart_state = state->state = DEVST_DHCPDISC;
-	} else if (dev->caps & CAP_DHCP) {
+	else if (dev->caps & CAP_DHCP)
 		state->restart_state = state->state = DEVST_DHCPREQ;
-	} else if (dev->caps & CAP_BOOTP) {
+	else if (dev->caps & CAP_BOOTP)
 		state->restart_state = state->state = DEVST_BOOTP;
-	}
 
 	state->next = slist;
 	slist = state;
 
+	n_devices++;
+
 	return 0;
 }
 
-static void parse_addr(uint32_t * addr, const char *ip)
+static void parse_addr(uint32_t *addr, const char *ip)
 {
 	struct in_addr in;
 	if (inet_aton(ip, &in) == 0) {
@@ -471,16 +576,16 @@ static unsigned int parse_proto(const char *ip)
 		fprintf(stderr, "%s: invalid protocol '%s'\n", progname, ip);
 		longjmp(abort_buf, 1);
 	}
-      bail:
+bail:
 	return caps;
 }
 
 static int add_all_devices(struct netdev *template);
 
-static int parse_device(struct netdev *dev, const char *ip)
+static int parse_device(struct netdev *dev, char *ip)
 {
 	char *cp;
-	int i, opt;
+	int opt;
 	int is_ip = 0;
 
 	dprintf("IP-Config: parse_device: \"%s\"\n", ip);
@@ -502,7 +607,7 @@ static int parse_device(struct netdev *dev, const char *ip)
 			dev->name = ip;
 		}
 	} else {
-		for (i = opt = 0; ip && *ip; ip = cp, opt++) {
+		for (opt = 0; ip && *ip; ip = cp, opt++) {
 			if ((cp = strchr(ip, ':'))) {
 				*cp++ = '\0';
 			}
@@ -533,7 +638,7 @@ static int parse_device(struct netdev *dev, const char *ip)
 			case 4:
 				strncpy(dev->hostname, ip, SYS_NMLN - 1);
 				dev->hostname[SYS_NMLN - 1] = '\0';
-				memcpy(dev->reqhostname, dev->hostname, 
+				memcpy(dev->reqhostname, dev->hostname,
 				       SYS_NMLN);
 				break;
 			case 5:
@@ -558,9 +663,10 @@ static int parse_device(struct netdev *dev, const char *ip)
 static void bringup_device(struct netdev *dev)
 {
 	if (netdev_up(dev) == 0) {
-		if (dev->caps) {
+		if (dev->caps)
 			add_one_dev(dev);
-		} else {
+		else {
+			dev->proto = PROTO_NONE;
 			complete_device(dev);
 		}
 	}
@@ -589,7 +695,7 @@ static void bringup_one_dev(struct netdev *template, struct netdev *dev)
 	bringup_device(dev);
 }
 
-static struct netdev *add_device(const char *info)
+static struct netdev *add_device(char *info)
 {
 	struct netdev *dev;
 	int i;
@@ -612,6 +718,9 @@ static struct netdev *add_device(const char *info)
 	if (bootp_init_if(dev) == -1)
 		goto bail;
 
+	if (packet_open(dev) == -1)
+		goto bail;
+
 	printf("IP-Config: %s hardware address", dev->name);
 	for (i = 0; i < dev->hwlen; i++)
 		printf("%c%02x", i == 0 ? ' ' : ':', dev->hwaddr[i]);
@@ -620,7 +729,7 @@ static struct netdev *add_device(const char *info)
 	       dev->caps & CAP_BOOTP ? " BOOTP" : "",
 	       dev->caps & CAP_RARP ? " RARP" : "");
 	return dev;
-      bail:
+bail:
 	free(dev);
 	return NULL;
 }
@@ -668,7 +777,8 @@ static int add_all_devices(struct netdev *template)
 		    (flags & (IFF_BROADCAST | IFF_POINTOPOINT))) {
 			dprintf("Trying to bring up %s\n", de->d_name);
 
-			if (!(dev = add_device(de->d_name)))
+			dev = add_device(de->d_name);
+			if (!dev)
 				continue;
 			bringup_one_dev(template, dev);
 		}
@@ -721,7 +831,7 @@ int ipconfig_main(int argc, char *argv[])
 {
 	struct netdev *dev;
 	int c, port;
-	int err;
+	int err = 0;
 
 	/* If progname is set we're invoked from another program */
 	if (!progname) {
@@ -799,8 +909,8 @@ int ipconfig_main(int argc, char *argv[])
 			       "dest to %d\n",
 			       cfg_local_port, cfg_remote_port);
 		}
-		loop();
+		err = loop();
 	}
 
-	return 0;
+	return err;
 }
